"""Annotate family

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

distribued with GPL-2 or later license

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""
from ..i18n import _
from ..error import DictConsistencyError
from ..utils import normalize_family
from .variable import Walk


#mode order is important
modes_level = ('basic', 'normal', 'expert')


class Mode:  # pylint: disable=R0903
    """Class to manage mode level
    """
    def __init__(self,
                 name: str,
                 level: int,
                 ) -> None:
        self.name = name
        self.level = level

    def __gt__(self,
               other: int,
               ) -> bool:
        return other.level < self.level


modes = {name: Mode(name, idx) for idx, name in enumerate(modes_level)}


class FamilyAnnotator(Walk):
    """Annotate family
    """
    def __init__(self,
                 objectspace,
                 ):
        self.objectspace = objectspace
        if not hasattr(self.objectspace.space, 'variables'):
            return
        self.remove_empty_families()
        self.family_names()
        self.change_modes()
        self.dynamic_families()
        self.convert_help()

    def _has_variable(self,
                      family: 'self.objectspace.family',
                      ) -> bool:
        if hasattr(family, 'variable'):
            for variable in family.variable.values():
                if isinstance(variable, self.objectspace.family):
                    if self._has_variable(variable):
                        return True
                else:
                    return True
        return False

    def remove_empty_families(self) -> None:
        """Remove all families without any variable
        """
        for families in self.objectspace.space.variables.values():
            removed_families = []
            for family_name, family in families.variable.items():
                if isinstance(family, self.objectspace.family) and not self._has_variable(family):
                    removed_families.append(family_name)
            for family_name in removed_families:
                del families.variable[family_name]

    def family_names(self) -> None:
        """Set doc, path, ... to family
        """
        for family in self.get_families():
            if not hasattr(family, 'description'):
                family.description = family.name
            family.doc = family.description
            del family.description
            family.name = normalize_family(family.name)

    def change_modes(self):
        """change the mode of variables
        """
        # default is high level
        new_family_mode = modes_level[-1]
        for family in self.objectspace.space.variables.values():
            self.change_family_mode(family,
                                    new_family_mode,
                                    first=True,
                                    )

    def change_family_mode(self,
                           family: 'self.objectspace.family',
                           new_family_mode: str,
                           first: bool=False,
                           ) -> str:
        if hasattr(family, 'mode'):
            family_mode = family.mode
        else:
            family_mode = new_family_mode
        if hasattr(family, 'variable'):
            # change variable mode, but not if variables are not in a family
            for variable in family.variable.values():
                if isinstance(variable, self.objectspace.family):
                    new_family_mode = self.change_family_mode(variable,
                                                              new_family_mode,
                                                              )
                    continue
                if first:
                    continue
                if isinstance(variable, self.objectspace.leadership):
                    func = self._change_variable_mode_leader
                else:
                    func = self._change_variabe_mode
                func(variable,
                     family_mode,
                     )
                if modes[new_family_mode] > modes[variable.mode]:
                    new_family_mode = variable.mode
        if not first:
            # set the lower variable mode to family
            family.mode = new_family_mode
        return new_family_mode

    def _change_variabe_mode(self,
                             variable,
                             family_mode: str,
                             is_follower=False,
                             ) -> None:
        # auto_save or auto_freeze variable is set to 'basic' mode
        # if its mode is not defined by the user
        if 'mode' not in vars(variable) and \
                (variable.auto_save is True or variable.auto_freeze is True):
            variable.mode = modes_level[0]
        # mandatory variable without value is a basic variable
        if variable.mandatory is True and not hasattr(variable, 'default'):
            variable.mode = modes_level[0]
        # none basic variable in high level family has to be in high level
        if modes[variable.mode] < modes[family_mode] and \
                (not is_follower or variable.mode != modes_level[0]):
            variable.mode = family_mode

    def _change_variable_mode_leader(self,
                                     leadership,
                                     family_mode: str,
                                     ) -> None:
        is_follower = False
        leader_mode = None
        for follower in leadership.variable:
            if follower.auto_save is True:
                msg = _(f'leader/followers "{follower.name}" could not be auto_save')
                raise DictConsistencyError(msg, 29, leadership.xmlfiles)
            if follower.auto_freeze is True:
                msg = f'leader/followers "{follower.name}" could not be auto_freeze'
                raise DictConsistencyError(_(msg), 30, leadership.xmlfiles)
            self._change_variabe_mode(follower,
                                      family_mode,
                                      is_follower,
                                      )
            if leader_mode is None:
                leader_mode = leadership.variable[0].mode
                leadership.variable[0].mode = None
            else:
                # leader's mode is minimum level
                if modes[leader_mode] > modes[follower.mode]:
                    follower.mode = leader_mode
            is_follower = True
        leadership.mode = leader_mode

    def dynamic_families(self):
        """link dynamic families to object
        """
        for family in self.get_families():
            if 'dynamic' not in vars(family):
                continue
            family.suffixes = self.objectspace.paths.get_variable(family.dynamic)
            del family.dynamic
            if not family.suffixes.multi:
                msg = _(f'dynamic family "{family.name}" must be linked '
                        f'to multi variable')
                raise DictConsistencyError(msg, 16, family.xmlfiles)
            for variable in family.variable.values():
                if isinstance(variable, self.objectspace.family):
                    msg = _(f'dynamic family "{family.name}" cannot contains another family')
                    raise DictConsistencyError(msg, 22, family.xmlfiles)

    def convert_help(self):
        """Convert variable help
        """
        for family in self.get_families():
            if not hasattr(family, 'help'):
                continue
            if not hasattr(family, 'information'):
                family.information = self.objectspace.information(family.xmlfiles)
            family.information.help = family.help
            del family.help
